#!/usr/bin/env python
# -*- coding: utf-8 -*-
#       shader.py
#
#       Copyright 2008 Gordon Allott <gordallott@gmail.com>
#
#       This program is free software; you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation; either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program; if not, write to the Free Software
#       Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#       MA 02110-1301, USA.
#
#       #
from render import *
from OpenGL.GL import *
from OpenGL.GLU import *
from OpenGL.GLUT import *
from OpenGL.GL.EXT.framebuffer_object import *
from OpenGL.GL.ARB.multitexture import *
from OpenGL.GL.ARB.shader_objects import *
from OpenGL.GL.ARB.fragment_shader import *
from OpenGL.GL.ARB.vertex_shader import *


shader_frag_bloom = """
uniform sampler2D srcSampler;
uniform sampler2D blurSampler;
uniform sampler2D grainSampler;
uniform float     sharpness;
uniform float     brightness;
uniform vec2      grain_mod;

#define SIGMOIDAL_BASE          2.0
#define SIGMOIDAL_RANGE         20.0

#define DO_BLOOM    1
#define DO_GRAIN    1
#define DO_SKETCH   1

void main() {
    vec4 blurcolor  = texture2D( blurSampler, gl_TexCoord[0].xy);
    vec4 basecolor  = texture2D( srcSampler, gl_TexCoord[0].xy);
    vec4 aux = texture2D( grainSampler, gl_TexCoord[0].xy * vec2(4.0, 2.0) + grain_mod);
    float graincolor = aux.a;
    vec3 sketchcolor = aux.rgb * 0.5;
    vec4 val = basecolor;
    vec4 bloom = vec4(0.0, 0.0, 0.0, 0.0);
    vec4 screen;


    if (DO_SKETCH) {
        // do sketch

        vec3 saturatedScene = clamp((val - 0.1) * 1.5, 0.0, 1.0);

        vec3 negativeSketch = val * (1 - sketchcolor);

        float sketchResult = (negativeSketch.r + negativeSketch.g + negativeSketch.b) * 0.33333;
        sketchResult = clamp(0.5 + sketchResult, 0.0, 1.0);

        val = vec4(saturatedScene * vec3(sketchResult), 1.0);

    }


    if (DO_BLOOM || DO_GRAIN) {
        // screen
        screen = 1.0 - ((1.0 - basecolor) * (1.0 - blurcolor));
    }

    if (DO_BLOOM) {

        bloom = -(SIGMOIDAL_BASE + (sharpness * SIGMOIDAL_RANGE)) * (blurcolor - 0.5);
        bloom = 1.0 + pow(vec4(2.718281828459045), bloom);
        bloom = 1.0 / bloom;
        bloom = bloom * brightness;



        val = 1.0 - ((1.0 - val) * (1.0 - bloom));


    }

    if (DO_GRAIN) {
        //soft light
        // (((1.0-L)*U*L)+ (L*R_s))
        //val = ((1.0 - basecolor) * graincolor * basecolor) + (basecolor * screen);
    }

    gl_FragColor = val;


}
"""

shader_frag_blur = """
uniform sampler2D srcSampler;
uniform vec2 blurSize;

#define NK7_0 0.19230769230769229
#define NK7_1 0.18269230769230768
#define NK7_2 0.15384615384615385
#define NK7_3 0.067307692307692304

vec4 GaussPass(sampler2D src, vec2 coord, vec2 blrsize) {
    //blrsize is the size (in texture coordinates) of the blur kernel

    vec4 accum;
    vec2 step, pos;
    step = blrsize / 7.0;
    pos = coord - (step * 3.0);

    accum  = texture2D(src, pos) * NK7_3; pos += step;
    accum += texture2D(src, pos) * NK7_2; pos += step;
    accum += texture2D(src, pos) * NK7_1; pos += step;
    accum += texture2D(src, pos) * NK7_0; pos += step;
    accum += texture2D(src, pos) * NK7_1; pos += step;
    accum += texture2D(src, pos) * NK7_2; pos += step;
    accum += texture2D(src, pos) * NK7_3; pos += step;

    return accum;

}

void main() {
    gl_FragColor = GaussPass(srcSampler, gl_TexCoord[0].xy, blurSize);
}
"""


shader_frag_bloom_opt = """
uniform sampler2D srcSampler;
uniform sampler2D blurSampler;
uniform sampler2D grainSampler;
uniform float     sharpness;
uniform float     brightness;

#define SIGMOIDAL_BASE          2.0
#define SIGMOIDAL_RANGE         20.0

void main() {
    vec4 blurcolor  = texture2D( blurSampler, gl_TexCoord[0].xy);
    vec4 basecolor  = texture2D( srcSampler, gl_TexCoord[0].xy);
    vec4 aux = texture2D(grainSampler, gl_TexCoord[1].xy);
    vec4 val = basecolor;
    vec4 bloom = vec4(0.0, 0.0, 0.0, 0.0);


    vec3 saturatedScene = clamp((val - 0.1) * 1.5, 0.0, 1.0);
    vec3 negativeSketch = val * (1 - aux.rgb * 0.5);
    float sketchResult = (negativeSketch.r + negativeSketch.g + negativeSketch.b) * 0.33333;
    sketchResult = clamp(0.5 + sketchResult, 0.0, 1.0);

    bloom = -(SIGMOIDAL_BASE + (sharpness * SIGMOIDAL_RANGE)) * (blurcolor - 0.5);
    bloom = (1.0 / (1.0 + pow(vec4(2.718281828459045), bloom))) * brightness;

    val = 1.0 - ((1.0 - vec4(saturatedScene * vec3(sketchResult), 1.0)) * (1.0 - bloom));

    gl_FragColor = val;


}
"""

shader_vert_default = """
uniform vec2      grain_mod;
void main() {
    gl_TexCoord[0] = gl_MultiTexCoord0;
    gl_TexCoord[1] = gl_MultiTexCoord0;
    gl_TexCoord[1].xy *= vec2(4.0, 2.0);
    gl_TexCoord[1].xy += grain_mod;
    gl_Position = gl_ModelViewProjectionMatrix * gl_Vertex;
}
"""

class Shader:

    def __init__(self, vert_shader="", frag_shader=""):
        if not glInitShaderObjectsARB():
            raise RuntimeError(
                """ARB Shader Objects extension is required """
                """but not supported on this machine!"""
            )

        self.program = self.compile_program(vert_shader, frag_shader)

    def compile_shader(self, source, shader_type):
        shader = glCreateShaderObjectARB(shader_type)
        glShaderSource(shader, source)
        #print shader
        glCompileShader(shader)

        #status = 1
        #status = glGetShaderiv(shader, GL_COMPILE_STATUS, status) #GL_FRAMEBUFFER_EXT)
        #print status
        #if not status:
            #self.print_log(shader)
            #glDeleteShader(shader)
            #raise ValueError, 'Shader compilation failed'



        return shader

    def compile_program(self, vertex_source, fragment_source):
        vertex_shader = None
        fragment_shader = None
        program = glCreateProgram()

        if vertex_source:
            vertex_shader = self.compile_shader(vertex_source, GL_VERTEX_SHADER)
            glAttachShader(program, vertex_shader)

        if fragment_source:
            fragment_shader = self.compile_shader(fragment_source, GL_FRAGMENT_SHADER)
            glAttachShader(program, fragment_shader)

        glLinkProgram(program)

        if vertex_shader:
            glDeleteShader(vertex_shader)

        if fragment_shader:
            glDeleteShader(fragment_shader)

        return program


    def print_log(shader):
        length = glGetShaderiv(shader, GL_INFO_LOG_LENGTH)

        if length > 0:
            log = ""
            glGetShaderInfoLog(shader, length, length, log)
            print log


    def bind(self):
        glUseProgram(self.program)

    def unbind(self):
        glUseProgram(0)




class PostProcess_Blur:

    fb = None
    shader = None
    render = None

    use_first = True

    draw_tex = 0

    def __init__(self, renderobj, size=(128, 128)):
        self.render = renderobj

        self.fb = Framebuffer(self.render, size, smooth=True, backbuffer=True)
        self.size = size

        self.shader = Shader(vert_shader=shader_vert_default,
                             frag_shader=shader_frag_blur)

        self.draw_tex = self.fb.get_texture()




    def process(self, input_tex):
        # first we draw the input buffer into the first framebuffer
        self.fb.bind()

        prev_tex = self.draw_tex
        if self.use_first:
            self.fb.flip()
            self.use_first = False
        else:
            self.use_first = True


        self.render.disable(GL_DEPTH_TEST)


        self.render.enable(GL_TEXTURE_2D)
        self.render.bind_texture(input_tex)
        self.render.enable(GL_BLEND)
        self.render.clearscreen()
        self.render.set_blend('addictive')
        self.render.set_color(0.5, 0.5, 0.5, 1.0)

        self.set_ortho()
        glLoadIdentity()
        self.draw_quad()

        self.render.bind_texture(prev_tex)#self.fb.get_texture())
        self.draw_quad()

        #self.fb.unbind()
        #self.shader.unbind()
        ##return the final texture
        #return self.fb.get_texture()


        #store the texture we got from that render.
        tex = self.fb.get_texture()


        #flip the buffers, do the first gaussian pass
        self.fb.flip()
        self.render.disable(GL_BLEND)
        self.render.set_color(1.0, 1.0, 1.0, 1.0)
        #assign the first fragment shader
        self.shader.bind()

        # set our variables
        loc = glGetUniformLocation(self.shader.program, 'srcSampler')
        glUniform1i(loc, 0)

        loc = glGetUniformLocation(self.shader.program, 'blurSize')
        glUniform2f(loc, 1.0 / self.size[0] * 7.0, 0.0)

        #draw the first pass into the back buffer
        self.render.bind_texture(tex)
        self.draw_quad()

        #store the texture we got from that render.
        tex = self.fb.get_texture()

        #flip the buffers, do the second gaussian pass
        self.fb.flip()

        #assign the second fragment shader
        self.shader.bind()

        # set our variables
        loc = glGetUniformLocation(self.shader.program, 'srcSampler')
        glUniform1i(loc, 0)

        loc = glGetUniformLocation(self.shader.program, 'blurSize')
        glUniform2f(loc, 0.0, 1.0 / self.size[1] * 7.0)

        # draw the second pass into the front buffer
        self.render.bind_texture(tex)
        self.draw_quad()

        self.fb.unbind()
        self.shader.unbind()
        #return the final texture
        self.draw_tex = self.fb.get_texture()
        return self.fb.get_texture()


    def set_ortho(self):
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()

        glOrtho(0.0, self.size[0], self.size[1], 0.0, -1.0, 1.0)

        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()



    def draw_quad(self, vmod=0.0, hmod=0.0):

        vmod = (1.0 / self.size[1]) * vmod
        hmod = (1.0 / self.size[1]) * hmod

        glLoadIdentity()

        glBegin(GL_QUADS)               # Start Drawing The face

        # Front Face (note that the texture's corners have to match the quad's corners)
        glTexCoord2f(0.0+hmod, 1.0+vmod); glVertex3f(0.0, 0.0, -1.0)
        glTexCoord2f(1.0+hmod, 1.0+vmod); glVertex3f(self.size[0], 0.0, -1.0)
        glTexCoord2f(1.0+hmod, 0.0+vmod); glVertex3f(self.size[0], self.size[1], -1.0)
        glTexCoord2f(0.0+hmod, 0.0+vmod); glVertex3f(0.0, self.size[1], -1.0)

        glEnd()


class PostProcess_Bloom:

    shader = None
    new_pos = 0

    def __init__(self, renderobj):
        self.render = renderobj
        glColor3f(1.0, 1.0, 1.0)
        self.shader = Shader(vert_shader=shader_vert_default,
                             frag_shader=shader_frag_bloom_opt)

        self.shader.bind()

        self.grain = self.render.create_texture('grain.png', repeats=('repeat', 'repeat'))

        # set our variables
        loc = glGetUniformLocation(self.shader.program, 'srcSampler')
        glUniform1i(loc, 0)

        loc = glGetUniformLocation(self.shader.program, 'blurSampler')
        glUniform1i(loc, 1)

        loc = glGetUniformLocation(self.shader.program, 'grainSampler')
        glUniform1i(loc, 2)

        loc = glGetUniformLocation(self.shader.program, 'sharpness')
        glUniform1f(loc, 0.4)

        loc = glGetUniformLocation(self.shader.program, 'brightness')
        glUniform1f(loc, 0.4)



        self.shader.unbind()




    def process(self, input_tex, blur_tex):

        self.render.disable(GL_DEPTH_TEST)
        self.render.disable(GL_BLEND)

        self.render.enable(GL_TEXTURE_2D)
        self.render.bind_texture(input_tex, blur_tex, self.grain)

        self.shader.bind()

        if self.new_pos > 1:

            loc = glGetUniformLocation(self.shader.program, 'grain_mod')
            glUniform2f(loc, random.gauss(0.0, 1.0), random.gauss(0.0, 1.0))

            self.new_pos = 0

        self.new_pos += 1

        try:
            glPushMatrix(GL_PROJECTION)
        except:
            glPushMatrix()
        self.set_ortho()

        self.draw_quad()

        try:
            glPopMatrix(GL_PROJECTION)
        except:
            glPopMatrix()
        self.shader.unbind()
        self.render.disable(GL_TEXTURE_2D)


    def set_ortho(self):
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()

        glOrtho(0.0, 1.0, 1.0, 0.0, -1.0, 1.0)

        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()



    def draw_quad(self):

        width = 1.0
        height = 1.0

        glBegin(GL_QUADS)               # Start Drawing The face

        # Front Face (note that the texture's corners have to match the quad's corners)
        glMultiTexCoord2fARB(GL_TEXTURE0, 0.0, 1.0); glMultiTexCoord2fARB(GL_TEXTURE1, 0.0, 1.0); glVertex3f(0.0, 0.0, -1.0);
        glMultiTexCoord2fARB(GL_TEXTURE0, 1.0, 1.0); glMultiTexCoord2fARB(GL_TEXTURE1, 1.0, 1.0); glVertex3f(width, 0.0, -1.0);
        glMultiTexCoord2fARB(GL_TEXTURE0, 1.0, 0.0); glMultiTexCoord2fARB(GL_TEXTURE1, 1.0, 0.0); glVertex3f(width,  height, -1.0);
        glMultiTexCoord2fARB(GL_TEXTURE0, 0.0, 0.0); glMultiTexCoord2fARB(GL_TEXTURE1, 0.0, 0.0); glVertex3f(0.0,  height, -1.0);

        glEnd();

        glActiveTextureARB(GL_TEXTURE0);


class Framebuffer:
    """ a framebuffer object """

    id = 0
    depth = 0
    texture = None
    backtexture = None
    size = (0, 0)
    currentTexture = 0 # 0 = front 1 = back

    def __init__(self, renderobj, size=(1, 1), smooth=False, backbuffer=False):

        if not  glInitFramebufferObjectEXT():
            raise RuntimeError(
                """EXT Framebuffer Objects extension is required """
                """but not supported on this machine!"""
            )

        self.id = glGenFramebuffersEXT(1)
        width, height = size
        self.size = size
        self.render = renderobj

        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, self.id)
        # -->
        # create our depth buffer


        self.depth = glGenRenderbuffersEXT(1)
        glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, self.depth)
        glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT, GL_DEPTH_COMPONENT24, width, height)

        glFramebufferRenderbufferEXT(
            GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_RENDERBUFFER_EXT,
            self.depth
        )

        self.texture = Texture(self.render, size=size, data='0', smooth=smooth, mipmap=False, repeats=('clamp', 'clamp'))

        glFramebufferTexture2DEXT(
            GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D,
            self.texture.get_id(),
            0 # mipmap level, normally 0
        )

        if backbuffer:
            self.backtexture = Texture(self.render, size=size, data='0', smooth=smooth, mipmap=False, repeats=('clamp', 'clamp'))
            glFramebufferTexture2DEXT(
                GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT1_EXT, GL_TEXTURE_2D,
                self.backtexture.get_id(),
                0 # mipmap level, normally 0
            )

        # check that it was created properly
        status = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT)

        if status == GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT_EXT:
            print 'FRAMEBUFFER_INCOMPLETE_ATTACHMENT_EXT'

        elif status == GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT_EXT:
            print 'FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT_EXT'

        elif status == GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS_EXT:
            print 'FRAMEBUFFER_INCOMPLETE_DIMENSIONS_EXT'

        elif status == GL_FRAMEBUFFER_INCOMPLETE_FORMATS_EXT:
            print 'FRAMEBUFFER_INCOMPLETE_FORMATS_EXT'

        elif status == GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER_EXT:
            print 'FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER_EXT'

        elif status == GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER_EXT:
            print 'FRAMEBUFFER_INCOMPLETE_READ_BUFFER_EXT'

        elif status == GL_FRAMEBUFFER_UNSUPPORTED_EXT:
            print 'FRAMEBUFFER_UNSUPPORTED_EXT'


        assert status == GL_FRAMEBUFFER_COMPLETE_EXT, status
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0)

        # <--

    def bind(self):
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, self.id)
        glDrawBuffer(GL_COLOR_ATTACHMENT0_EXT)
        self.currentTexture = 0
        glPushAttrib(GL_VIEWPORT_BIT)
        self.set_projection()

    def flip(self):
        if not self.backtexture:
            return

        if self.currentTexture:
            glDrawBuffer(GL_COLOR_ATTACHMENT0_EXT)
            self.currentTexture = 0
        else:
            glDrawBuffer(GL_COLOR_ATTACHMENT1_EXT)
            self.currentTexture = 1

    def get_texture(self):
        if not self.backtexture:
            return self.texture

        if self.currentTexture:
            return self.backtexture
        else:
            return self.texture

    def unbind(self):
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0)
        glPopAttrib()


    def set_projection(self, depth=100.0):

        width, height = self.size

        height = float(height) if height != 0 else 1.0
        width = float(width)

        glViewport(0, 0, int(width), int(height))

        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        glOrtho(0.0, 1.0, 0.0, 1.0, -1.0, 1.0)
        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()

    def clear(self):
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0)


    def test_fb(self):

        self.render.bind_texture(self.get_texture())
        self.render.enable(GL_TEXTURE_2D)
        self.render.disable(GL_BLEND)
        self.render.set_color(1.0, 1.0, 1.0)

        #glColor4f(1.0, 0.0, 1.0, 1.0)

        self.set_projection()
        glLoadIdentity()
        width = 1.0
        height = 1.0

        width, height = self.size

        hmod, vmod = (0.0, 0.0)
        glBegin(GL_QUADS)               # Start Drawing The face

        glTexCoord2f(0.0+hmod, 1.0+vmod); glVertex3f(0.0, 0.0, 0.0)
        glTexCoord2f(1.0+hmod, 1.0+vmod); glVertex3f(1.0, 0.0, 0.0)
        glTexCoord2f(1.0+hmod, 0.0+vmod); glVertex3f(1.0, 1.0, 0.0)
        glTexCoord2f(0.0+hmod, 0.0+vmod); glVertex3f(0.0, 1.0, 0.0)

        glEnd()


