#!/usr/bin/env python
# -*- coding: utf-8 -*-
#       actors.py
#
#       Copyright 2008 Gordon Allott <gordallott@gmail.com>
#
#       This program is free software; you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation; either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program; if not, write to the Free Software
#       Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#       MA 02110-1301, USA.
#
#
import ode
from ode import Infinity as INFINITY
import gamelib.instance as instance

import physsim
import gamelib.render
from math import *
import math
from props import drawable
from gamelib.timeline import Timeline, time_now, time_start
from gamelib.util import cosine_interpolate, linear_interpolate
import random
import copy

def cubicInterpolate(v0, v1, v2, v3, x):
    """ performs generic cubic interpolation """
    P = (v3 - v2) - (v0 - v1)
    Q = (v0 - v1) - P
    R = v2 - v0
    S = v1

    return P * math.pow(x, 3) + Q * math.pow(x, 2) + R * x + S

#------------------------------------------------------------------------------
# Actors
#------------------------------------------------------------------------------

def findid(store, id):
    related_actor = None
    for actor in store.itervalues():
        if hasattr(actor, 'logic_id'):
            if actor.logic_id == id:
                related_actor = actor
                break

    if related_actor == None:
        raise Exception, ("can't find actor", id)

    return related_actor

class Actor(physsim.DynamicBodyUpright, drawable):

    dynamics_bound = 'capsule'

    dyn_material = { 'bounce':0.0, 'min_bounce':0.1, 'friction':INFINITY, 'density':50  }
    repeats = ('clamp', 'clamp')

    def __init__(self, gameworld, renderlib, pos):
        self.gameworld = gameworld
        physsim.DynamicBodyUpright.__init__(self, pos=pos)
        drawable.__init__(self, renderlib)


    def move(self, x=0.0, y=0.0):

        mass = self.dynamics_body.getMass().mass * 10

        x *= mass
        y *= mass
        self.dynamics_body.addForce((x, y, 0.0))

    def create_self(self, *args, **kwargs):
        physsim.DynamicBodyUpright.create_self(self, *args, **kwargs)

        # Make our feet rays
        self.dynamics_ray = ode.GeomRay(self.dynamics_space, (self.dynamics_size[1] / 2.0) + 0.2)
        self.dynamics_ray.set((0, 0, 0), (0, -1, 0))
        self.dynamics_ray.super_object = self

        self.dynamics_rayl = ode.GeomRay(self.dynamics_space, (self.dynamics_size[1] / 2.0) + 0.2 )
        self.dynamics_rayl.set((0, 0, 0), (-0.2, -1, 0))
        self.dynamics_rayl.super_object = self

        self.dynamics_rayr = ode.GeomRay(self.dynamics_space, (self.dynamics_size[1] / 2.0) + 0.2 )
        self.dynamics_rayr.set((0, 0, 0), (0.2, -1, 0))
        self.dynamics_rayr.super_object = self

    def step(self):
        self.is_colliding = False
        physsim.DynamicBodyUpright.step(self)

        # update the position of the feet ray
        self.dynamics_ray.set(self.dynamics_body.getPosition(), (0, -1, 0))
        self.dynamics_rayl.set(self.dynamics_body.getPosition(), (-0.2, -1, 0))
        self.dynamics_rayr.set(self.dynamics_body.getPosition(), (0.2, -1, 0))
        # reset the collision from for the feet ray


class Level_End(physsim.DynamicObject):
    """ an actor, when you collide with this it triggers off an end level signal
    """
    dyn_material = {'bounce': 0.0, 'min_bounce': 0.0, 'friction': INFINITY, 'density':INFINITY }
    step_me = False
    dynamics_size = (4.0, 4.0)
    collide = False
    current_sprite = 'foobar'

    def draw(self, *args, **kwargs):
        pass

    def set_flip(self, *args, **kwargs):
        pass

    def __init__(self, gameworld, renderlib, pos):
        self.gameworld = gameworld
        physsim.DynamicObject.__init__(self, pos=pos)

        self.connect('collide', self.cb_collide)

    def cb_collide(self, ref, object, info):
        if object.isplayer:
            self.gameworld.end_level = True


class Death_Area(physsim.DynamicObject):
    """ an actor, counts its self as harmful
    """
    dyn_material = {'bounce': 0.0, 'min_bounce': 0.0, 'friction': INFINITY, 'density':INFINITY }
    step_me = False
    dynamics_size = (4.0, 4.0)
    collide = False
    current_sprite = 'foobar'
    isharmful = True

    def draw(self, *args, **kwargs):
        pass

    def set_flip(self, *args, **kwargs):
        pass

    def __init__(self, gameworld, renderlib, pos):
        self.gameworld = gameworld
        physsim.DynamicObject.__init__(self, pos=pos)
        self.isharmful = True



class Death_Area_Trigger(Death_Area):
    """ only turns into a death area on trigger (and removes its death area on
    'close' trigger)
    """

    def __init__(self, gameworld, renderlib, pos, id='noid'):
        Death_Area.__init__(self, gameworld, renderlib, pos)
        self.logic_id = id
        self.isharmful = False

    def action(self, status=True):
        self.isharmful = not status


class Warp_Start(physsim.DynamicObject):
    """ an actor, when you collide with this it triggers off an end level signal
    """
    dyn_material = {'bounce': 0.0, 'min_bounce': 0.0, 'friction': INFINITY, 'density':INFINITY }
    step_me = False
    dynamics_size = (1.0, 1.0)
    collide = False
    current_sprite = 'foobar'
    sound = None

    def draw(self, *args, **kwargs):
        pass

    def set_flip(self, *args, **kwargs):
        pass

    def __init__(self, gameworld, renderlib, pos, id='noid', out_id='warpend'):
        self.gameworld = gameworld
        physsim.DynamicObject.__init__(self, pos=pos)

        self.out_id = out_id.split(',')
        self.current_warp = 0
        self.logic_id = id

        self.connect('collide', self.cb_collide)
        self.sound = instance.instance.audio.create_sound('teleport.ogg')


    def cb_collide(self, ref, object, info):
        if object.isplayer:
            id = self.out_id[self.current_warp]
            endpos = findid(self.gameworld.actors, id)
            endpos = endpos.get_position()
            endpos = (endpos[0], endpos[1] + 1.0)

            object.post_move = endpos
            object.post_velocity = (0.0, 4.0)

            endpos = (endpos[0], endpos[1]-0.5)
            self.light, self.light_id = self.gameworld.add_light(position=endpos, color=(1.0, 1.0, 1.0),
                                              texture='portal', size=(1.5, 1.5), return_id=True, dynamic=True)

            object.timeline.add_action(time=1.0, relative=time_now, repeat_callback=False, callback=self.kill_effects)

            instance.instance.audio.play(self.sound)

    def action(self, status):
        if status:
            self.current_warp += 1
            if self.current_warp >= len(self.out_id):
                self.current_warp = 0

    def kill_effects(self, progress):
        if hasattr(self, 'light'):
            del self.light
            del self.gameworld.lights_dynamic[self.light_id]
            del self.light_id
        #del self.light

class Warp_End(physsim.DynamicObject):
    """ an actor, when you collide with this it triggers off an end level signal
    """
    dyn_material = {'bounce': 0.0, 'min_bounce': 0.0, 'friction': INFINITY, 'density':INFINITY }
    step_me = False
    dynamics_size = (1.0, 1.0)
    collide = False
    current_sprite = 'foobar'

    def draw(self, *args, **kwargs):
        pass

    def set_flip(self, *args, **kwargs):
        pass

    def __init__(self, gameworld, renderlib, pos, id='warpend'):
        self.gameworld = gameworld
        physsim.DynamicObject.__init__(self, pos=pos)

        self.logic_id = id

class LightProxy(physsim.DynamicObject):
    """ an actor, when you collide with this it triggers off an end level signal
    """
    dyn_material = {'bounce': 0.0, 'min_bounce': 0.0, 'friction': INFINITY, 'density':INFINITY }
    step_me = True
    dynamics_size = (1.0, 1.0)
    collide = False
    current_sprite = 'foobar'

    on_color = (0.0, 1.0, 0.0, 1.0)
    off_color = (1.0, 0.0, 0.0, 1.0)


    def draw(self, *args, **kwargs):
        pass

    def set_flip(self, *args, **kwargs):
        pass

    def __init__(self, gameworld, renderlib, pos, id='lightpuzzle', fadeout=5.0):
        self.gameworld = gameworld
        physsim.DynamicObject.__init__(self, pos=pos)
        self.logic_id = id
        self.fadeout = float(fadeout)
        self.current_fade = self.fadeout

    def action(self, status=True):
        if status:
            self.current_fade = self.fadeout

    def step(self):
        self.current_fade -= 1.0 / 30.0
        fade = self.current_fade / self.fadeout
        self.gameworld.light_ambiant = (fade, fade, fade)

class MultiSwitch(physsim.DynamicObject):
    """ recieves multiple switches """
    dyn_material = {'bounce': 0.0, 'min_bounce': 0.0, 'friction': INFINITY, 'density':INFINITY }
    step_me = True
    dynamics_size = (1.0, 1.0)
    graphics_size = (1.0, 1.0)

    current_sprite = 'foobar'
    wait = 0.0

    timeline = None
    collide = False
    switch_on = False

    sounds = {'dingding.ogg': None,
              '32937__digifishmusic__Grandfather_Clock_digifish.ogg': None,
             }

    playing_wait_sound = False

    related_ids = []

    def __init__(self, gameworld, renderlib, pos, id='multiswitch', related_id='NoID',
                 input_ids='NoID', auto_close_wait=0.0):
        self.gameworld = gameworld
        physsim.DynamicObject.__init__(self, pos=pos)
        #drawable.__init__(self, renderlib)

        if related_id == 'NoID':
            raise Exception, 'No related ids given to a switch'
        if input_ids == 'NoID':
            raise Exception, "No inputs given"

        self.related_ids = related_id.split(',')
        self.related_actors = []
        #for rel_id in self.related_ids:
            #self.related_actors.append(findid(self.gameworld.actors, rel_id))

        self.input_actors = []
        self.input_ids = input_ids.split(',')
        #for input_id in input_ids:
            ##print 'Multiswitch',id,'input Ids',findid(self.gameworld.actors, input_id)
            #self.input_actors.append(findid(self.gameworld.actors, input_id))

        self.base_pos = pos
        self.timeline = Timeline()
        self.logic_id = id

        self.auto_close_wait = float(auto_close_wait)

        self.do_switch = False
        self.switch_on = False

        for filename in self.sounds:
            self.sounds[filename] = instance.instance.audio.create_sound(filename)

    def _find_actors(self):
        for rel_id in self.related_ids:
            self.related_actors.append(findid(self.gameworld.actors, rel_id))
        for input_id in self.input_ids:
            #print 'Multiswitch',id,'input Ids',findid(self.gameworld.actors, input_id)
            self.input_actors.append(findid(self.gameworld.actors, input_id))

    def action(self, status=True):
        if len(self.input_actors) == 0:
            self._find_actors()
        if status:
            switch_test = True
            for test_input in self.input_actors:
                switch_test = test_input.switch_on
                if not test_input.switch_on:
                    break
            if switch_test is not self.switch_on:
                self.switch_on = True
                self.do_switch = True
        else:
            if self.switch_on is not status:
                self.switch_on = False
                self.do_switch = True

    def turn_off(self, progress):
        if progress >= 1.0:
            if self.switch_on:
                self.switch_on = False
                self.do_switch = True

    def draw(self, *args, **kwargs):
        """ """

    def set_flip(self, *args, **kwargs):
        """ """

    def step(self):

        if self.do_switch:
            if self.switch_on:
                instance.instance.audio.play(self.sounds['dingding.ogg'])

            for related_actor in self.related_actors:
                related_actor.action(self.switch_on)

            self.do_switch = False

        self.timeline.step()



class LightPuzzle_A_Solver(physsim.DynamicObject):
    """ an actor, when you collide with this it triggers off an end level signal
    """
    dyn_material = {'bounce': 0.0, 'min_bounce': 0.0, 'friction': INFINITY, 'density':INFINITY }
    step_me = False
    dynamics_size = (1.0, 1.0)
    collide = False
    current_sprite = 'foobar'

    on_color = (0.0, 1.0, 0.0, 1.0)
    off_color = (1.0, 0.0, 0.0, 1.0)

    def draw(self, *args, **kwargs):
        pass

    def set_flip(self, *args, **kwargs):
        pass

    def __init__(self, gameworld, renderlib, pos, id='lightpuzzle', watch_ids='noid', final_ids='noid'):
        self.gameworld = gameworld
        physsim.DynamicObject.__init__(self, pos=pos)
        self.logic_id = id

        self.watch_ids = watch_ids.split(',')
        self.final_ids = final_ids.split(',')

        self.solved = False

        self.sound = instance.instance.audio.create_sound('solved.ogg')

    def action(self, status=True):
        if not self.solved:
            watch_actors = []
            for id in self.watch_ids:
                watch_actors.append(findid(self.gameworld.actors, id))

            all_activated = True
            for actor in watch_actors:
                if not actor.activated:
                    all_activated = False

            if all_activated:
                trigger_actors = []
                for id in self.final_ids:
                    trigger_actors.append(findid(self.gameworld.actors, id))

                for actor in trigger_actors:
                    actor.action(True)

                self.solved = True
                instance.instance.audio.play(self.sound)

class LightPuzzle_A(physsim.DynamicObject):
    """ an actor, when you collide with this it triggers off an end level signal
    """
    save_me = ['activated', 'light_color']
    dyn_material = {'bounce': 0.0, 'min_bounce': 0.0, 'friction': INFINITY, 'density':INFINITY }
    step_me = False
    dynamics_size = (1.0, 1.0)
    collide = False
    current_sprite = 'foobar'

    on_color = (1.0, 0.6, 0.2, 1.0)#ff9933
    off_color = (0.2, 0.6, 1.0, 1.0)#3399ff

    def draw(self, *args, **kwargs):
        pass

    def set_flip(self, *args, **kwargs):
        pass

    def __init__(self, gameworld, renderlib, pos, id='lightpuzzle', activated=False):
        self.gameworld = gameworld
        physsim.DynamicObject.__init__(self, pos=pos)
        self.logic_id = id
        self.activated = bool(activated)

        self.light = self.gameworld.add_light(position=pos, color=(1.0, 0.0, 0.0),
                                              texture='light_generic.png', size=(3.0, 3.0), dynamic=True)

        if self.activated:
            self.light.color = self.on_color

        else:
            self.light.color = self.off_color


    def action(self, status=True):
        if status:
            if self.activated:
                self.activated = False
                self.light.color = self.off_color
            else:
                self.light.color = self.on_color
                self.activated = True

        #solver = findid(self.gameworld.actors, self.solver_id)
        #solver.action(status=True)


class GhostPoint(physsim.DynamicObject, drawable):
    dyn_material = {'bounce': 0.0, 'min_bounce': 0.0, 'friction': INFINITY, 'density':INFINITY }
    step_me = True
    dynamics_size = (0.5, 0.5)
    graphics_size = (0.5, 0.5)
    collide = False
    current_sprite = 'foobar'
    sound = None

    graphics_filenames = ['sprites/ghost_point.png']
    current_sprite = 'sprites/ghost_point.png'

    def __init__(self, gameworld, renderlib, pos, time=3.0, ghosts=2):
        self.gameworld = gameworld
        drawable.__init__(self, renderlib)
        physsim.DynamicObject.__init__(self, pos=pos)

        self.logic_id = id
        self.max_time = float(time)
        self.watched_threads = 0

        self.connect('collide', self.cb_collide)
        self.sound = instance.instance.audio.create_sound('teleport.ogg')

        self.timeline = Timeline()
        self.busy = False
        self.let_collide = True

    def cb_collide(self, ref, object, info):
        def reset():
            self.timeline.add_action(time=0.0, relative=time_now,
                                    repeat_callback=True, length=1.0,
                                    callback=self._reset_point)
            return False

        if self.let_collide and object.isplayer and not getattr(object,'isghost',False):
            #self.timeline.add_action(time=0.0, relative=time_now,
                                     #repeat_callback=True, length=self.max_time+1,
                                     #callback=self.end_thread)
            self.watch_progress = self.max_time/(self.max_time+1.0)
            object.max_length = self.max_time
            self.player_obj = object
            self.let_collide = False

            # If this hasn't watched any threads and there are copies running around, ignore this attempt
            if (object.copies[1]['copy'] is not None or object.copies[2]['copy'] is not None) and self.watched_threads == 0:
                return reset()

            # If there is anything in the action records, clear it
            if self.watched_threads == 0:
                object.copies[1]['action_record'] = []
                object.copies[2]['action_record'] = []

            # Increment the thread
            self.watched_threads += 1

            # If we reach the thread limit, reset to 0 and return
            if self.watched_threads > 2:
                self.watched_threads = 0
                return reset()
            # Start recording the thread
            self.player_obj.start_rec_thread(self.watched_threads)
            self.timeline.add_action(time=0.0, relative=time_now,
                                     repeat_callback=True, length=self.max_time+1,
                                     callback=self.end_thread)
            self.busy = True

    def _reset_point(self, progress):
        if progress >= 1.0:
            self.let_collide = True

    def end_thread(self, progress):
        #print progress, self.watch_progress
        if progress >= 1.0:
            self.let_collide = True
        elif progress >= self.watch_progress and self.busy:
            self.busy = False
            for thread_index in range(1, self.watched_threads+1):
                thread = self.player_obj.copies[thread_index]
                # If its still recording, stop it
                if thread['do_record']:
                    self.player_obj.stop_rec_thread(thread_index)
                # Make sure the copies are cleaned up
                if self.player_obj.copies[thread_index]['copy'] is not None:
                    self.player_obj.copies[thread_index]['copy'].cleanup()
                # Spawn the thread
                self.player_obj.spawn_thread(thread_index)
            # Teleport the player back the the thread point
            to_pos = self.get_position()
            self.player_obj.move_to(to_pos[0], to_pos[1], (0,0))
        else:
            self.gameworld.draw_fs(progress, 'addictive')

    def step(self):
        self.timeline.step()

class Switch(physsim.DynamicObject, drawable):
    """ A basic actor, designed to just move up and down in an attempt to block
    to players path unless they get their timing right

     we don't use the Actor object becase we don't need to be upright or even
     have a ode body
    """
    graphics_filenames = ['sprites/switch.png', 'sprites/switch_on.png']
    current_sprite = 'sprites/switch.png'
    dyn_material = {'bounce': 0.0, 'min_bounce': 0.0, 'friction': INFINITY, 'density':INFINITY }
    step_me = True
    dynamics_size = (1.0, 1.0)
    graphics_size = (1.0, 1.0)

    wait = 0.0

    timeline = None
    collide = False
    switch_on = False

    sounds = {'dingding.ogg': None,
              '32937__digifishmusic__Grandfather_Clock_digifish.ogg': None,
             }

    playing_wait_sound = False

    related_ids = []

    def __init__(self, gameworld, renderlib, pos, id='NoID', related_id='NoID',
                 auto_close_wait=0.0, kill_id='NoID', key_id='NoID'):
        self.gameworld = gameworld
        physsim.DynamicObject.__init__(self, pos=pos)
        drawable.__init__(self, renderlib)

        if related_id == 'NoID' and kill_id == 'NoID':
            raise Exception, 'No related ids given to a switch'
        if id != 'NoID':
            self.logic_id = id

        self.kill = []
        self.kill_id = kill_id.split(',')

        self.key_objects = []
        self.key_id = key_id.split(',')

        self.related_ids = related_id.split(',')

        #for index in range(len(self.related_ids)):
            #self.related_ids[index] = float(self.related_ids[index]

        self.base_pos = pos
        self.timeline = Timeline()

        self.action = False

        self.auto_close_wait = float(auto_close_wait)

        self.connect('collide', self.cb_collide)
        self.do_switch = False

        for filename in self.sounds:
            self.sounds[filename] = instance.instance.audio.create_sound(filename)

    def cb_collide(self, ref, object, info):
        self._search_obj()
        if object.isplayer or object in self.key_objects:
            if not self.switch_on:
                self.switch_on = True
                self.fire_events()
                self.current_sprite = 'sprites/switch_on.png'

                if self.auto_close_wait > 0.0:
                    self.timeline.add_action(time=self.auto_close_wait, relative=time_now,
                                             repeat_callback=False, callback=self.turn_off)
                    self.playing_wait_sound = instance.instance.audio.play(self.sounds['32937__digifishmusic__Grandfather_Clock_digifish.ogg'])

    def turn_off(self, progress):
        self.switch_on = False
        self.fire_events()
        if self.playing_wait_sound:
            self.playing_wait_sound.stop(fadeout=500)
            self.playing_wait_sound = None

        self.current_sprite = 'sprites/switch.png'

    def _search_obj(self):
        if self.kill_id[0] != 'NoID' and len(self.kill) == 0:
            for kill_obj in self.kill_id:
                self.kill.append(findid(self.gameworld.actors, kill_obj))
        if self.key_id[0] != 'NoID' and len(self.key_objects) == 0:
            for key_obj in self.key_id:
                try:
                    obj = findid(self.gameworld.actors, key_obj)
                except:
                    obj = findid(self.gameworld.props, key_obj)
                self.key_objects.append(obj)

    def fire_events(self):
        #find our related_ids
        related_actors = []
        for related_id in self.related_ids:
            if related_id == 'NoID':
                break
            related_actor = None
            for actor in self.gameworld.actors.itervalues():
                if hasattr(actor, 'logic_id'):
                    if actor.logic_id == related_id:
                        related_actor = actor
                        break

            if related_actor == None:
                raise Exception, "can't find actor %s"%related_id

            related_actors.append(related_actor)

        self.related_actors = related_actors
        self.do_switch = True

    def step(self):

        if self.do_switch:
            if self.switch_on:
                instance.instance.audio.play(self.sounds['dingding.ogg'])
                for kill_obj in self.kill:
                    kill_obj.disable_me = True
            else:
                for kill_obj in self.kill:
                    kill_obj.disable_me = False

            for related_actor in self.related_actors:
                related_actor.action(self.switch_on)

            self.do_switch = False

        self.timeline.step()

class Area_Switch(Switch):
    """ an actor, when you collide with this it triggers off an end level signal
    """

    def draw(self, *args, **kwargs):
        pass

    def set_flip(self, *args, **kwargs):
        pass

    def cb_collide(self, ref, object, info):
        if object.isplayer:
            if not self.switch_on:
                self.switch_on = True
                self.fire_events()
                self.current_sprite = 'sprites/switch_on.png'

                if self.auto_close_wait > 0.0:
                    self.timeline.add_action(time=self.auto_close_wait, relative=time_now,
                                             repeat_callback=False, callback=self.turn_off)
                    self.playing_wait_sound = instance.instance.audio.play(self.sounds['32937__digifishmusic__Grandfather_Clock_digifish.ogg'])

    def __init__(self, *args, **kwargs):
        Switch.__init__(self, *args, **kwargs)

        if kwargs.has_key('key_id'):
            self.key_id = key_id



class Door(physsim.DynamicObject, drawable):
    """ A basic actor, designed to just move up and down in an attempt to block
    to players path unless they get their timing right

     we don't use the Actor object becase we don't need to be upright or even
     have a ode body
    """
    graphics_filenames = ['sprites/door/door_open.png',
                          'sprites/door/door_close1.png',
                          'sprites/door/door_close2.png']
    #graphics_filenames = ['sprites/door/door_open.png']
    current_sprite = 'sprites/door/door_open.png'
    dyn_material = {'bounce': 0.0, 'min_bounce': 0.0, 'friction': INFINITY, 'density':INFINITY }
    step_me = True
    dynamics_size = (1.0, 4.0)
    graphics_size = (2.0, 4.0)

    timeline = None
    status_closed = True
    do_switch = False

    repeats = ('clamp', 'clamp')

    def __init__(self, gameworld, renderlib, pos, id='NoID', move_pos=2.0, arch=True):
        self.gameworld = gameworld

        physsim.DynamicObject.__init__(self, pos=pos)
        drawable.__init__(self, renderlib)

        self.logic_id = id
        self.base_pos = pos
        self.timeline = Timeline()
        self.do_switch = False
        self.status_closed = True
        self._status_check = True
        self.move_pos = float(move_pos)

        self.schedule_anim = True

        if arch:
            self.gameworld.add_geom(size=self.graphics_size, position=self.base_pos,
                               graphic_filename='sprites/door_front.png',
                               no_collide=True,
                               no_detect=True,
                               draw_infront=True,
                               raw_filename=True)

            self.gameworld.add_geom(size=self.graphics_size, position=self.base_pos,
                               graphic_filename='sprites/door_back.png',
                               no_collide=True,
                               no_detect=True,
                               draw_infront=False,
                               raw_filename=True)

        self.sound = instance.instance.audio.create_sound('zing.ogg')


    def action(self, status=True):
        self.do_switch = True

    def open(self, progress):
        if self.current_sprite == 'sprites/door_h.png':
            self.current_sprite = 'sprites/door_h'

        self.status_closed = False
        c = cosine_interpolate(1.0, 0.0, progress)
        self.set_color((1.0, 1.0, 1.0, c))

        self.dynamics_geom.setCategoryBits(0)
        self.dynamics_geom.setCollideBits(0)

        if progress >= 1.0:
            self.status_closed = False

    def close(self, progress):
        self.status_closed = True
        c = cosine_interpolate(0.0, 1.0, progress)
        self.set_color((1.0, 1.0, 1.0, c))
        #self.set_position(self.base_pos[0], pos)
        if progress >= 1.0:
            self.status_closed = True

    def animate(self, progress):
        if 'sprites/door/door_open.png' not in self.graphics_filenames:
            return
        if progress == 1.0:
            self.current_sprite = 'sprites/door/door_open.png'
            self.schedule_anim = True
        elif progress >= 0.6:
            self.current_sprite = 'sprites/door/door_close2.png'
        else:
            self.current_sprite = 'sprites/door/door_close1.png'

    def step(self):

        if self.do_switch:
            if self.status_closed:
                self.timeline.add_action(time=0.0, relative=time_now,
                                         repeat_callback=True, length=3.0,
                                         callback=self.open)

                if not self.graphics_sprites[self.current_sprite].check_bounds():
                        instance.instance.audio.play(self.sound)

                self.dynamics_geom.setCategoryBits(0)
                self.dynamics_geom.setCollideBits(0)

            else:
                self.timeline.add_action(time=0.0, relative=time_now,
                                         repeat_callback=True, length=0.3,
                                         callback=self.close)

                self.dynamics_geom.setCategoryBits(self.catagory)
                self.dynamics_geom.setCollideBits(self.collides)
            self.do_switch = False

        elif self.status_closed and self.dynamics_geom.getCategoryBits() == 0:
            self.dynamics_geom.setCategoryBits(self.catagory)
            self.dynamics_geom.setCollideBits(self.collides)
            self.set_color((1.0, 1.0, 1.0, 1.0))

        if self.schedule_anim:
            self.schedule_anim = False
            self.timeline.add_action(time=random.randint(3,10), relative=time_now,
                                         repeat_callback=True, length=0.1,
                                         callback=self.animate)
        self.timeline.step()

class Door_h(Door):
    graphics_filenames = ['sprites/door_h.png',
                          'sprites/door_h']
    current_sprite = 'sprites/door_h.png'

    dynamics_size = (4.0, 1.0)
    graphics_size = (4.0, 1.0)

    repeats = ('clamp', 'clamp')

    def __init__(self, *args, **kwargs):
        kwargs['arch'] = False
        Door.__init__(self, *args, **kwargs)


class UppetyBox(physsim.DynamicObject, drawable):
    """ A basic actor, designed to just move up and down in an attempt to block
    to players path unless they get their timing right

     we don't use the Actor object becase we don't need to be upright or even
     have a ode body
    """
    graphics_filenames = ['sprites/UppetyBox']
    current_sprite = 'sprites/UppetyBox'
    dyn_material = {'bounce': 0.0, 'min_bounce': 0.0, 'friction': INFINITY, 'density':INFINITY }
    step_me = True
    dynamics_size = (2.0, 2.0)
    graphics_size = (2.0, 2.0)

    timeline = None
    moving_down = True
    islift = True
    do_switch = False
    wait=0.0

    type = 'uppity_box'

    isharmful = True

    repeats = ('clamp', 'clamp')

    def __init__(self, gameworld, renderlib, pos, id='noid', timing=0.0, wait=0.0, move_to=-2.0, only_on_trigger=False):
        self.gameworld = gameworld
        physsim.DynamicObject.__init__(self, pos=pos)
        drawable.__init__(self, renderlib)

        timing = float(timing)

        self.base_pos = pos
        self.logic_id = id

        self.timeline = Timeline()
        self.only_on_trigger = bool(only_on_trigger)

        #switch action
        self.wait = float(wait)
        self.move_to = float(move_to)

        if not self.only_on_trigger:

            self.timeline.add_action(time=timing, relative=time_now,
                                         repeat_callback=False,
                                         callback=self.switch_action)

        self.sound = instance.instance.audio.create_sound('thud.ogg')

        self.disable_me = False

        #self.connect('collide', self.cb_collide)

    #def cb_collide(self, ref, object, info):
        #print object
        #if hasattr(object, 'type'):
            #print object
            #if object.type == 'prop_plank':
                #self.disable_me = True


    def action(self, status):
        if status:
            self.moving_down = False
            self.switch_action(1.0)

    def switch_action(self, progress):
        self.do_switch = True


    def smashy(self, progress):
        if not self.disable_me:
            top, bottom = (self.base_pos[1], self.base_pos[1] + self.move_to)

            #pos = cubicInterpolate(top+2.0, top, bottom, bottom-2.0, progress)
            pos = cosine_interpolate(top, bottom, progress)
            self.moving_down = True
            self.set_position(self.base_pos[0], pos)

            if progress >= 1.0:
                if not self.graphics_sprites[self.current_sprite].check_bounds():
                    instance.instance.audio.play(self.sound)

    def return_to_position(self, progress):
        if not self.disable_me:
            top, bottom = (self.base_pos[1], self.base_pos[1] + self.move_to)
            #pos = cubicInterpolate(bottom-2.0, bottom, top, top+2.0, progress)
            pos = cosine_interpolate(bottom, top, progress)
            self.moving_down = False
            self.set_position(self.base_pos[0], pos)


    def step(self):
        self.timeline.step()



        if self.do_switch and not self.disable_me:
            if self.moving_down:

                self.timeline.add_action(time=self.wait, relative=time_now,
                                         repeat_callback=True, length=3.0,
                                         callback=self.return_to_position)
                switch_in = self.wait + 3.0
            else:

                self.timeline.add_action(time=self.wait, relative=time_now,
                                         repeat_callback=True, length=0.6,
                                         callback=self.smashy)
                switch_in = self.wait + 0.3


            if not self.moving_down:

                self.timeline.add_action(time=switch_in, relative=time_now,
                                         repeat_callback=False,
                                         callback=self.switch_action)

            if self.moving_down and not self.only_on_trigger:
                self.timeline.add_action(time=switch_in, relative=time_now,
                                         repeat_callback=False,
                                         callback=self.switch_action)

            self.do_switch = False



class Spikes(physsim.DynamicObject, drawable):
    """ A basic actor, designed to just move up and down in an attempt to block
    to players path unless they get their timing right

     we don't use the Actor object becase we don't need to be upright or even
     have a ode body
    """
    graphics_filenames = ['sprites/spikes.png']
    current_sprite = 'sprites/spikes.png'
    dyn_material = {'bounce': 0.0, 'min_bounce': 0.0, 'friction': INFINITY, 'density':INFINITY }
    step_me = False
    dynamics_size = (2.0, 0.5)
    graphics_size = (2.0, 0.5)

    isharmful = True

    repeats = ('clamp', 'clamp')

    def __init__(self, gameworld, renderlib, pos):
        self.gameworld = gameworld
        physsim.DynamicObject.__init__(self, pos=pos)
        drawable.__init__(self, renderlib)


#------------------------------------------------------------------------------
# Player
#------------------------------------------------------------------------------


class Player(Actor):
    graphics_filenames = ['sprites/main_char.png',
                          'sprites/walk_cycle_final',
                          'sprites/climb',
                          'sprites/duck',
                          'sprites/jump',
                          'sprites/fall.png']
    current_sprite = 'sprites/main_char.png'
    dynamics_size = (0.8, 1.27)
    graphics_size = (1.5, 1.5)

    type = 'player_actor'
    isplayer = True

    run_speed = 8.0
    climb_speed = 5.0
    max_linear_move = 4.0

    can_jump = True

    post_move = False
    post_velocity = False

    dyn_material = { 'bounce': 0.0, 'min_bounce':0.1, 'friction':10024, 'density':5000  }

    sounds = {'jump1.ogg': None,
              'fall.ogg': None,
              'hurt.ogg': None,
             }


    touching_rope = False
    _temp_touch_rope = False
    touching_rope_do_animation = False
    do_jump_animation = False
    on_lift = False #if this is true we ignore jumping animations for the moment

    timeline = None
    is_hurt = False

    last_player = None
    max_length = 7.0 # The number of seconds a thread can last for

    def __init__(self, *args, **kwargs):
        Actor.__init__(self, *args, **kwargs)

        self.connect('collide', self.cb_collide)

        for filename in self.sounds:
            self.sounds[filename] = instance.instance.audio.create_sound(filename)

        self.timeline = Timeline()
        self.is_hurt = False

        if self.last_player is not None:
            for thread in self.last_player.copies:
                if self.last_player.copies[thread]['copy'] is not None:
                    self.last_player.copies[thread]['copy'].cleanup()
        self.last_player = self

        #self.copies = [] # copies spawned by the player
        #self.action_record = {} # records what the player is doing in the format: action_record[frame] = {pos, vel, sprite_name, teleport=bool}
        #self.current_frame = 1 # for time keeping purposes

        self.record_flip_x = False
        self.record_flip_y = False

        self.copies = {}
        self.copies[1] = { "do_record":False, "action_record":[], "copy":None, "allow_spawn":True }
        self.copies[2] = { "do_record":False, "action_record":[], "copy":None, "allow_spawn":True }

    def cb_clear_jump(self, progress):
        if progress == 1.0:
            self.can_jump = True
            return

        self.can_jump = False

    def cb_collide(self, ref, object, info):

        if object.isharmful:
            instance.instance.audio.play(self.sounds['hurt.ogg'])
            self.is_hurt = True

        if (object.current_sprite == 'geoms/rope.png' or object.current_sprite == 'geoms/rope_start.png') or\
           (object.current_sprite == 'geoms/rope_2.png' or object.current_sprite == 'geoms/rope_2_start.png'):
            self._temp_touch_rope = object
            return

        if object.islift:
            self.on_lift = True
            return


        if not self.on_lift and object.collide and not object.isplayer:
            instance.instance.audio.play(self.sounds['fall.ogg'])



    def move(self, x=0.0, y=0.0):
        if abs(self.dynamics_body.getLinearVel()[0]) < self.max_linear_move:
            Actor.move(self, x, y)

    def go_left(self):
        self.set_flip(False, False)

        c_speed = self.dynamics_body.getLinearVel()[0]
        if c_speed > -self.run_speed:
            velocity=(-(c_speed - -self.run_speed), 0.0)
        else:
            velocity = (0.0, 0.0)

        if not self.is_colliding:
            velocity = (-0.5, 0.0) # for jumping

        self.move(velocity[0], velocity[1])

    def go_right(self, slow=False):
        self.set_flip(True, False)
        c_speed = self.dynamics_body.getLinearVel()[0]
        if c_speed <self.run_speed:
            velocity=((self.run_speed - c_speed), 0.0)
        else:
            velocity = (0.0, 0.0)

        if not self.is_colliding:
            velocity = (0.5, 0.0) # for jumping

        if slow:
            velocity = (velocity[0] * 0.25, velocity[1] * 0.25)

        self.move(velocity[0], velocity[1])

    def set_flip(self, x=False, y=False):
        Actor.set_flip(self, x, y)
        self.record_flip_x = x
        self.record_flip_y = y

    def jump(self):

        #        sound = self.audio.create_sound('jump1.ogg')
        #self.audio.play(sound)

        #get our instance


        velocity=(0.0, 0.0)
        check_jump = True

        if abs(self.dynamics_body.getLinearVel()[0]) < self.max_linear_move:
            if self.touching_rope:

                check_jump = False

                setPos = self.touching_rope.dynamics_geom.getPosition()
                selfPos = self.dynamics_body.getPosition()
                distance = abs(setPos[0] - selfPos[0])

                if distance == 0.0:

                    c_speed = self.dynamics_body.getLinearVel()[1]
                    if c_speed < self.climb_speed:
                        velocity=(0.0, (self.climb_speed - c_speed))

                    self.touching_rope_do_animation = True

                elif distance < 0.3:
                    self.move_to( setPos[0], selfPos[1] )

                else:
                    check_jump = True

            elif check_jump:
                if not self.is_colliding:
                    velocity = (0.0, 0.5) # for jumping
                else:
                    if self.can_jump:
                        instance.instance.audio.play(self.sounds['jump1.ogg'])
                        self.do_jump_animation = True
                        velocity=(0.0, 14.0)

                        self.timeline.add_action(time=0.0, relative=time_now,
                                             repeat_callback=True, length=0.3,
                                             callback=self.cb_clear_jump)

            self.move(velocity[0], velocity[1])

    def move_to(self, x, y, velocity=(0,0) ):
        """ Move the player to a different point, for use in portals """
        self.dynamics_body.disable()
        self.dynamics_body.setPosition( (x, y, 0) )
        self.dynamics_body.enable()
        self.dynamics_body.setLinearVel( (velocity[0], velocity[1], 0) )

    def set_graphics(self, velocity):

        if self.touching_rope and self.touching_rope_do_animation:
            self.current_sprite = 'sprites/climb'
            return

        if not self.on_lift:
            if velocity[1] > 0.08:
                self.current_sprite = 'sprites/jump'
                if self.do_jump_animation:
                    self.graphics_sprites[self.current_sprite].texture.repeat = False
                    self.graphics_sprites[self.current_sprite].texture.current_frame = 0
                return


            if velocity[1] < -0.08:
                self.current_sprite = 'sprites/fall.png'
                return

        if abs(velocity[0]) > 0.1:
            self.current_sprite = 'sprites/walk_cycle_final'
            return


        self.current_sprite = 'sprites/main_char.png'

    def step(self):
        Actor.step(self)


        linear_vel = self.dynamics_body.getLinearVel()

        if abs(linear_vel[1]) > 0.25:
            self.can_jump = False
        else:
            self.can_jump = True

        self.set_graphics(linear_vel)

        self.touching_rope_do_animation = False
        self.do_jump_animation = False
        self._temp_touch_rope = False
        self.on_lift = False

        self.timeline.step()

        #if linear_vel[1] > 0.08:
            #self.set_sprite('jump')

        #if linear_vel[1] < -0.08:
            #self.set_sprite('fall')

    def clear_record(self):
        for copy in self.copies:
            copy.cleanup()
        del self.action_record

    # modified Simulator.add_object to account for this
    def post_step(self):
        """ Do something after wards """
        #print self.touching_rope, self._temp_touch_rope
        if self.touching_rope != self._temp_touch_rope:
            self.touching_rope = self._temp_touch_rope

        # For teleportation
        if self.post_move:
            self.move_to(self.post_move[0], self.post_move[1], velocity=self.post_velocity)
            self.post_move = False
            self.post_velocity = False

        # For recording the player
        if self.copies[1]['do_record'] or self.copies[2]['do_record']:
            record_entry = {"pos":self.dynamics_body.getPosition(), "vel":self.dynamics_body.getLinearVel(),
                            "sprite_name": self.current_sprite, "flip_x":self.record_flip_x,
                            "flip_y":self.record_flip_y}
            if self.copies[1]['do_record']:
                self.copies[1]['action_record'].append(record_entry)
            if self.copies[2]['do_record']:
                self.copies[2]['action_record'].append(record_entry)

            self.record_flip_x = False
            self.record_flip_y = False

    def thread_status(self, progress, thread=None):
        if not thread:
            return False
        if self.copies[thread]['do_record']:
            self.copies[thread]['status'].position = self.get_position()
            self.copies[thread]['status'].flip = self.graphics_sprites[self.current_sprite].flip
            c = cosine_interpolate(0.0, 1.0, 1.0-progress)
            self.copies[thread]['status'].color = (c, c, c, 1.0)
        if progress >= 1.0:
            if self.copies[thread]['do_record']:
                self.stop_rec_thread(thread)

    def toggle_thread(self, thread=1):
        if self.copies[thread]['do_record']:
            self.stop_rec_thread(thread)
        else:
            self.start_rec_thread(thread)

    def start_rec_thread(self, thread):
        self.copies[thread]['status_action'] = self.timeline.add_action(time=0.0, relative=time_now,
                                 repeat_callback=True, length=self.max_length,
                                 callback=self.thread_status, extra_args={'thread':thread})
        status, id = self.gameworld.add_light(color=(1.0, 1.0, 1.0),
                                   texture='thread', size=(2.0, 2.0),
                                   return_id=True, dynamic=True)
        self.copies[thread]['status'] = status
        self.copies[thread]['status_id']  = id
        self.copies[thread]['do_record'] = True
        self.copies[thread]['action_record'] = []

    def stop_rec_thread(self, thread):
        if self.copies[thread]['do_record']:
            self.copies[thread]['do_record'] = False
            self.timeline.remove_action(self.copies[thread]['status_action'], False)
            del self.copies[thread]['status']
            del self.gameworld.lights_dynamic[self.copies[thread]['status_id']]

    def spawn_thread(self, thread_num):
        """ spawn messr thread """
        thread = self.copies[thread_num]
        if thread['copy'] is None and not thread['do_record'] and len(thread['action_record']) > 0:
            new_copy, id = self.gameworld.add_actor(actor_object=PlayerCopy, return_id=True)
            new_copy.setup(thread['action_record'], self, thread_num, id)
        else:
            print "NOES",thread['copy'], thread['do_record'], len(thread['action_record'])

    #def set_thread(self, thread, object):
    #    self.copies[thread]['copy'] = object



#class PlayerCopy(physsim.DynamicObject, drawable):
class PlayerCopy(Actor):

    graphics_filenames = ['sprites/main_char.png',
                          'sprites/walk_cycle_final',
                          'sprites/climb',
                          'sprites/duck',
                          'sprites/jump',
                          'sprites/fall.png']
    current_sprite = 'sprites/main_char.png'
    dynamics_size = (0.8, 1.27)
    graphics_size = (1.5, 1.5)

    type = 'player_actor'
    isplayer = True
    isghost = True
    step_me = False

    catagory = physsim.CATAGORY_GENERIC
    collides = physsim.CATAGORY_GENERIC | physsim.CATAGORY_GEOM

    dyn_material = { 'bounce': 0.0, 'min_bounce':0.1, 'friction':10024, 'density':5000  }

    sounds = {'jump1.ogg': None,
              'fall.ogg': None,
              'hurt.ogg': None,
             }

    #def __init__(self, gameworld, renderlib, pos, *args, **kwargs):
        #self.gameworld = gameworld
        #physsim.DynamicObject.__init__(self, pos=pos)
        #drawable.__init__(self, renderlib)
    def __init__(self, *args, **kwargs):
        Actor.__init__(self, *args, **kwargs)

        self.start_playback = False
        self.record = {}
        self.current_frame = 1
        self.set_color( (1.0, 1.0, 1.0, 0.5) )

    def setup(self, record, parent_player, thread_id, id):
        """ Gives the copy the parameters needed for playback """
        self.my_id = id
        self.record = copy.copy(record)
        target_pos = self.record[0]['pos']
        self.set_position(target_pos[0], target_pos[1])
        self.start_playback = True
        self.player = parent_player
        self.thread_id = thread_id
        self.player.copies[self.thread_id]['copy'] = self

    def cleanup(self):
        """ Cleans up this copy """
        self.player.copies[self.thread_id]['copy'] = None
        self.gameworld.phys_world.remove_object(self.dynamics_id)
        del self.gameworld.actors[self.my_id]
        del self

    def step(self):
        pass

    def post_step(self):
        """ plays back a record of the player """
        #{"pos":self.dynamics_body.getPosition(), "vel":self.dynamics_body.getLinearVel(),
        #"sprite_name": self.current_sprite, "teleport":self.record_move}
        if self.start_playback:
            entry = self.record[self.current_frame]

            self.current_sprite = entry['sprite_name']
            self.set_flip(entry['flip_x'], entry['flip_y'])
            #self.dynamics_body.setLinearVel(entry['vel'])
            #if entry['teleport']:
                #self.dynamics_body.setPosition(entry['pos'])
            #self.dynamics_body.setPosition(entry['pos'])
            #speed = 0.0
            #if self.current_frame > 0:
            #    last_pos = self.record[self.current_frame - 1]['pos']
            #    speed = math.hypot(entry['pos'][0]-last_pos[0], entry['pos'][1]-last_pos[1])
            #if speed < and math.hypot(entry['vel'][0], entry['vel'][1]):
            #    print "Player trying to push something"
            #print math.hypot(entry['vel'][0], entry['vel'][1])
            self.dynamics_geom.setPosition(entry['pos'])
            next_frame = self.current_frame + 1
            if next_frame < len(self.record):
                self.current_frame += 1
            else:
                self.start_playback = False
                self.cleanup()
        else:
            print "Mwahaha"
