#!/usr/bin/env python
# -*- coding: utf-8 -*-
#       gameworld.py
#
#       Copyright 2008 Gordon Allott <gordallott@gmail.com>
#
#       This program is free software; you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation; either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program; if not, write to the Free Software
#       Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#       MA 02110-1301, USA.
#
#
import pygame
from pygame.locals import *
from random import randrange
import physsim
from props import Geom, Geom_nodetect
import gamelib.render
from ..util import make_id
from OpenGL.GL import *
import os, sys
import math
from gamelib.timeline import Timeline, time_now, time_start
from gamelib.util import hex_to_gl
from gamelib.util import cosine_interpolate, linear_interpolate

class Layer:

    layer_pos = None
    layer_movement = (1.0, 1.0)

    sprites = []

    def __init__(self, renderobj, pos=0, relative_movement=(1.0, 1.0)):
        """ creates a new layer,
        """
        self.render = renderobj
        self.layer_pos = pos
        self.layer_movement = relative_movement
        self.sprites = []

    def draw(self, current_translation):
        # set our relative movement
        x, y = self.layer_movement

        new_translation = list(current_translation)
        new_translation[0] *= x
        new_translation[1] *= y


        #self.render.reset()
        self.render.set_screen_pos(new_translation[0], new_translation[1])
        self.render.reset()
        self.render.set_screen()

        for sprite, movement in self.sprites:
            sprite.position = (sprite.position[0] + movement[0], sprite.position[1] + movement[1])
            if sprite.position[1] > 9980.0:
                sprite.position = (sprite.position[0], sprite.position[1] - 1000.0)
                sprite.draw()

            sprite.position = (sprite.position[0], sprite.position[1] % 1000.0)
            sprite.draw()

    def add_sprite(self, filename, size, position, movement=(0.0, 0.0)):
        filename = os.path.join('layered/', filename)

        tex = self.render.create_texture(filename, repeats=('clamp', 'clamp'))
        sprite = gamelib.render.Sprite(self.render, tex,
                                       size=size,
                                       tile_repeats=(1.0, 1.0))
        sprite.position = position

        self.sprites.append((sprite, movement))

    def add_graphic(self, filename, tex_size=(1.0, 1.0), height=10.0, y_position=0.0, movement=(0.0, 0.0), color=(1.0, 1.0, 1.0, 1.0)):

        repeats = list(tex_size)
        repeats[0] = 1000.0 / repeats[0]
        repeats[1] = height / repeats[1]

        repeat_mode = ['repeat', 'repeat']
        if repeats[0] == 1.0:
            repeat_mode[0] = 'clamp'

        if repeats[1] == 1.0:
            repeat_mode[1] = 'clamp'

        filename = os.path.join('layered/', filename)

        tex = self.render.create_texture(filename, repeats=repeat_mode)
        sprite = gamelib.render.Sprite(self.render, tex,
                                       size=(1000.0, height),
                                       tile_repeats=repeats)
        sprite.color = color
        sprite.position = (0.0, y_position)

        self.sprites.append((sprite, movement))

def cubicInterpolate(v0, v1, v2, v3, x):
    """ performs generic cubic interpolation """
    P = (v3 - v2) - (v0 - v1)
    Q = (v0 - v1) - P
    R = v2 - v0
    S = v1
    return P * math.pow(x, 3) + Q * math.pow(x, 2) + R * x + S

class world:

    props = {}
    geoms = {}
    geoms_decor_infront = {}
    geoms_decor_behind = {}
    actors = {}
    lights = {}
    layers = []
    misc = []

    timeline = None

    fade_blend = 'subtractive'
    fade_value = 1.0
    do_fade = False

    end_level = False

    #bg_gradient = ((0xba/255.0, 0x5f/255.0, 0x94/255.0), (0x9a/255.0, 0x87/255.0, 0xb4/255.0))
    bg_gradient = (hex_to_gl('#ba5f94'), hex_to_gl('#9a87b4'))
    #bg_gradient = ((1.0, 1.0, 1.0), (0.0, 0.0, 0.0))

    def __init__(self, renderlib):
        self.render = renderlib
        self.phys_world = physsim.Simulation()
        self.timeline = Timeline()
        self.end_level = False


        self.props = {}
        self.geoms = {}
        self.geoms_decor_infront = {}
        self.geoms_decor_behind = {}
        self.actors = {}
        self.lights = {}
        self.lights_noambiant = {}
        self.lights_dynamic = {}
        self.layers = []

        self.fade_blend = 'subtractive'
        self.fade_value = 1.0
        self.do_fade = False

        self.base_ambiant = (0.9, 0.9, 0.9)
        self.bg_ambiant = (1.0, 1.0, 1.0)
        self.light_ambiant = (1.0, 1.0, 1.0)

        self.light_list = None



    def shutdown(self):

        self.props = {}
        self.geoms = {}
        self.geoms_decor_infront = {}
        self.geoms_decor_behind = {}
        self.actors = {}
        self.lights = {}
        self.lights_noambiant = {}
        self.lights_dynamic = {}
        self.layers = []

        self.fade_blend = 'subtractive'
        self.fade_value = 1.0
        self.do_fade = False

        self.phys_world = None
        self.timeline = None


    def timeline_cb_fadein_black(self, position):
        #self.fade_value = cubicInterpolate(1.5, 1.0, 0.0, -0.5, position)

        self.fade_value = cosine_interpolate(1.0, 0.0, position)
        self.fade_blend = 'subtractive'
        self.do_fade = True



    def timeline_cb_fadein_white(self, position):
        self.fade_value = cosine_interpolate(1.0, 0.0, position)
        self.fade_blend = 'subtractive'
        self.do_fade = True

    def timeline_cb_fadeout_white(self, position):
        self.fade_value = cosine_interpolate(0.0, 1.0, position)
        self.fade_value = 1.0 - self.fade_value
        self.fade_blend = 'addictive'
        self.do_fade = True

    def add_misc(self, object):
        self.misc.append(object)


    def add_prop(self, prop_object=None, position=(0.0, 0.0), override_size=None, params={}):

        #filename = os.path.join('sprites/', graphic_filename)
        object = prop_object(self.render, position, override_size=override_size, **params)

        self.phys_world.add_object(object)
        object.set_position(position[0], position[1])


        id = make_id(store=self.props)
        self.props[id] = object

        return object

    def add_geom(self, size=(1.0, 1.0), position=(0.0, 0.0), graphic_filename='moo.png', raw_filename=False,
                 tiles=(1.0, 1.0), no_collide=False, no_detect=False, is_platform=False,
                 draw_infront=False):

        if raw_filename:
            filename = graphic_filename
        else:
            filename = os.path.join('geoms/', graphic_filename)

        geom_base_object = Geom
        if no_detect:
            geom_base_object = Geom_nodetect

        object = geom_base_object(self.render, position, size=size, graphics_filename=filename, tiles=tiles)

        if no_collide:
            object.collide = False

        if is_platform:
            object.isplatform = True

        if not no_detect:
            self.phys_world.add_object(object)

        object.set_position(position[0], position[1])


        store = self.geoms
        if no_detect:
            if draw_infront:
                store = self.geoms_decor_infront
            else:
                store = self.geoms_decor_behind

        grid_pos = (math.floor(position[0] * 0.1), math.floor(position[1] * 0.1))

        if not store.has_key(grid_pos):
            store[grid_pos] = [[], None]

        #id = make_id(store=store)
        #store[id] = object

        store[grid_pos][0].append(object)

        return object

    def optimise_geoms(self):
        """ creates display lists for our geoms """
        #glLoadIdentity()
        self.render.base_scalar = (1.0, 1.0, 1.0)
        self.render.base_rotation = (0.0, 0.0, 0.0)
        self.render.base_position = (0.0, 0.0, -1.0)

        for key, geomgrid in self.geoms.iteritems():
            self.render.reset()
            self.render.set_screen()

            new_list = gamelib.render.DisplayList()
            new_list.compile()

            for geom in geomgrid[0]:
                geom.draw(bounds_check=False, colors=False)

            new_list.end_compile()
            geomgrid[1] = new_list


    def optimise_back_geoms(self):
        self.render.base_scalar = (1.0, 1.0, 1.0)
        self.render.base_rotation = (0.0, 0.0, 0.0)
        self.render.base_position = (0.0, 0.0, -1.0)

        for key, geomgrid in self.geoms_decor_behind.iteritems():
            self.render.reset()
            self.render.set_screen()

            new_list = gamelib.render.DisplayList()
            new_list.compile()

            for geom in geomgrid[0]:
                geom.draw(bounds_check=False, colors=False)

            new_list.end_compile()
            geomgrid[1] = new_list


    def optimise_lights(self):

        self.render.base_scalar = (1.0, 1.0, 1.0)
        self.render.base_rotation = (0.0, 0.0, 0.0)
        self.render.base_position = (0.0, 0.0, -1.0)
        self.render.reset()
        self.render.set_screen()

        new_list = gamelib.render.DisplayList()
        new_list.compile()

        for light in self.lights.itervalues():
            light.draw()

        new_list.end_compile()

        self.light_list = new_list


    def add_actor(self, actor_object=None, position=(0.0, 0.0), params={}, force_size=None, return_id=False):

        object = actor_object(self, self.render, position, **params)
        if force_size:
            object.dynamics_size = force_size

        self.phys_world.add_object(object)
        object.set_position(position[0], position[1])


        id = make_id(store=self.actors)
        self.actors[id] = object

        if return_id:
            return object, id
        return object

    def save_state(self):
        """ saves the states of actors in the game world """
        saved_actors = {}
        for actor_id in self.actors:
            target = self.actors[actor_id]
            # Gather the timeline variables
            if hasattr(target, 'timeline') and not getattr(target, "isplayer", False):
                saved_actors[actor_id] = {}
                saved_actors[actor_id]['current_time'] = target.timeline.current_time
                saved_actors[actor_id]['actions'] = {}
                for id, action in target.timeline.actions.iteritems():
                    base_action = {}
                    base_action['start'] = action['start']
                    base_action['length'] = action['length']
                    base_action['repeating'] = action['repeating']
                    base_action['repeat_call'] = action['repeat_call']
                    base_action['repeat_wait'] = action['repeat_wait']
                    base_action['callback'] = action['callback']
                    base_action['iterations'] = action['iterations']
                    saved_actors[actor_id]['actions'][id] = base_action
            # Check if the actor wants to save any variables
            if hasattr(target, 'save_me') and not getattr(target, "isplayer", False):
                if not saved_actors.has_key(actor_id):
                    saved_actors[actor_id] = {}
                saved_actors[actor_id]['states'] = {}
                for attr in target.save_me:
                    saved_actors[actor_id]['states'][attr] = getattr(target, attr)

        return saved_actors

    def restore_state(self, state):
        """ restores the states of actors in the game world """
        for actor_id, timeline_state in state.iteritems():
            # Restore the timeline variables here
            if hasattr(self.actors[actor_id], "timeline"):
                self.actors[actor_id].timeline.current_time = timeline_state['current_time']
                for action_id, action in timeline_state['actions'].iteritems():
                    self.actors[actor_id].timeline.actions[action_id] = action
            # Restore the saved variables
            if timeline_state.has_key('states'):
                for attr in timeline_state['states']:
                    setattr(self.actors[actor_id], attr, timeline_state['states'][attr])

    def create_layer(self, pos=0, relative_movement=(1.0, 1.0)):
        newlayer = Layer(self.render, pos=pos, relative_movement=relative_movement)
        self.layers.append(newlayer)
        def layer_sort(x, y):
            return x.layer_pos - y.layer_pos

        self.layers.sort(layer_sort)
        return newlayer

    def add_light(self, position=(0.0, 0.0), color=(0.5, 0.5, 0.5), texture='light_generic.png',
                  size=(5.0, 5.0), return_id=False, ambiant_affected=True, dynamic=False):

        color = (color[0], color[1], color[2], 1.0)

        filename = os.path.join('lights/', texture)
        tex = self.render.create_texture(filename, repeats=('clamp', 'clamp'))


        sprite = gamelib.render.Sprite(self.render, tex, size=size, tile_repeats=(1.0, 1.0), blend='addictive')
        sprite.position = position

        sprite.color = color

        store = self.lights
        if not ambiant_affected:
            store = self.lights_noambiant

        if dynamic:
            store = self.lights_dynamic

        id = make_id(store=store)
        store[id] = sprite
        if return_id:
            return sprite, id


        return sprite

    def step(self):
        self.phys_world.step()
        self.timeline.step()
        for object in self.misc:
            object.step()

    def draw(self):
        #what we want to do is just get the 9 grid's around our current position
        current_grid = self.render.base_position
        current_grid = (-math.floor(current_grid[0] * 0.1), -math.floor(current_grid[1] * 0.1))

        self.draw_fs_grad()

        self.render.reset()
        self.render.set_screen()


        mainpos = self.render.base_position

        self.render.set_ambiant(*self.bg_ambiant)

        for layer in self.layers:
            layer.draw(mainpos)

        self.render.set_screen_pos(mainpos[0], mainpos[1])
        self.render.reset()
        self.render.set_screen()

        self.render.set_ambiant(*self.base_ambiant)

        #for geomgrid in self.geoms_decor_behind.itervalues():
            #for geom in geomgrid[0]:
                #geom.draw()

        for object in self.misc:
            object.draw()

        p_screen = self.render.base_position

        geomgrids = []
        for y in range(-2, 3):
            for x in range(-2, 3):
                tmp = (current_grid[0] + x, current_grid[1] + y)
                if self.geoms_decor_behind.has_key((tmp[0] ,tmp[1])):
                    geomgrids.append(self.geoms_decor_behind[tmp])

        self.render.reset()
        self.render.set_screen()
        self.render.set_color(1.0, 1.0, 1.0, 1.0)
        for geomgrid in geomgrids:
            if geomgrid[1] == None:
                for geom in geomgrid[0]:
                    geom.draw()

            else:
                self.render.reset()
                self.render.set_screen()
                geomgrid[1].execute()




        self.render.reset()
        self.render.base_position = p_screen
        self.render.set_screen()


        for prop in self.props.itervalues():
            prop.draw()


        p_screen = self.render.base_position

        geomgrids = []
        for y in range(-2, 3):
            for x in range(-2, 3):
                tmp = (current_grid[0] + x, current_grid[1] + y)
                if self.geoms.has_key((tmp[0] ,tmp[1])):
                    geomgrids.append(self.geoms[tmp])

        self.render.reset()
        self.render.set_screen()
        self.render.set_color(1.0, 1.0, 1.0, 1.0)
        for geomgrid in geomgrids:
            if geomgrid[1] == None:
                for geom in geomgrid[0]:
                    geom.draw()

            else:
                self.render.reset()
                self.render.set_screen()
                geomgrid[1].execute()

        self.render.reset()
        self.render.base_position = p_screen
        self.render.set_screen()


        self.render.set_ambiant(1.0, 1.0, 1.0)

        for light in self.lights_noambiant.itervalues():
            light.draw()


        self.render.set_ambiant(*self.base_ambiant)

        for actor in self.actors.itervalues():
            actor.draw()


        for geomgrid in self.geoms_decor_infront.itervalues():
            for geom in geomgrid[0]:
                geom.draw()

        self.render.set_ambiant(*self.light_ambiant)

        for light in self.lights_dynamic.itervalues():
                light.draw()

        if self.light_list:
            self.render.reset()
            self.render.set_screen()
            self.render.set_color(1.0, 1.0, 1.0, 1.0)
            self.light_list.execute()
        else:
            for light in self.lights.itervalues():
                light.draw()




        self.render.set_ambiant(1.0, 1.0, 1.0)

        #for light in self.lights_noambiant.itervalues():
            #light.draw()

        if self.do_fade:
            self.draw_fs(fade=self.fade_value, blend=self.fade_blend)

        self.render.reset()
        #self.render.set_screen()



    def draw_fs_grad(self):
        scale = self.render.base_scalar
        height = 10.0 * 1.0 / self.render.base_scalar[1]

        self.render.reset()
        self.render.disable(GL_BLEND)
        self.render.disable(GL_TEXTURE_2D)

        glBegin(GL_QUADS)

        glColor3fv(self.bg_gradient[0])
        glVertex3f(-100.0, -height, -1.0)
        glVertex3f( 100.0, -height, -1.0)

        glColor3fv(self.bg_gradient[1])
        glVertex3f( 100.0, height, -1.0)
        glVertex3f(-100.0, height, -1.0)

        glColor3fv((1.0, 1.0, 1.0))

        glEnd()


    def draw_fs(self, fade=1.0, blend='subtractive'):
        scale = self.render.base_scalar
        height = 10.0 * 1.0 / self.render.base_scalar[1]

        self.render.reset()
        self.render.enable(GL_BLEND)
        self.render.set_blend(blend)
        self.render.disable(GL_TEXTURE_2D)

        glBegin(GL_QUADS)

        glColor3fv((fade, fade, fade))
        glVertex3f(-100.0, -height, -1.0)
        glVertex3f( 100.0, -height, -1.0)
        glVertex3f( 100.0, height, -1.0)
        glVertex3f(-100.0, height, -1.0)

        glEnd()

