#!/usr/bin/env python
# -*- coding: utf-8 -*-
#       props.py
#
#       Copyright 2008 Gordon Allott <gordallott@gmail.com>
#
#       This program is free software; you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation; either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program; if not, write to the Free Software
#       Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#       MA 02110-1301, USA.
#
#       #
import ode
from ode import Infinity as INFINITY
import gamelib.instance as instance
from gamelib.timeline import Timeline, time_now, time_start

import physsim
import gamelib.render
from math import *

class drawable:

    graphics_sprites = {}
    current_sprite = 'moo.png'
    graphics_color = (1.0, 1.0, 1.0, 1.0)

    graphics_filenames = ['moo.png']
    graphics_size = (1.0, 1.0)
    graphics_tiles = (1.0, 1.0)
    graphics_textures = {}

    repeats = ('repeat', 'repeat')


    def __init__(self, renderlib):
        self.render = renderlib

        self.graphics_sprites = {}

        for filename in self.graphics_filenames:
            texture = self.render.create_texture(filename, repeats=self.repeats)

            sprite = gamelib.render.Sprite(self.render, texture,
                                           size=self.graphics_size, tile_repeats=self.graphics_tiles)

            self.graphics_sprites[filename] = sprite


    def set_flip(self, fliph=False, flipv=False):
        for sprite in self.graphics_sprites.itervalues():
            sprite.flip = (fliph, flipv)

    def set_color(self, color):
        for sprite in self.graphics_sprites.itervalues():
            sprite.color = color

    def set_sprite(self, filename):
        self.current_sprite = filename

    def get_sprite(self):
        return self.graphics_sprites[self.current_sprite]

    def draw(self, bounds_check=True, colors=True):
        """ draws the prop """
        pos = self.get_position()
        rot = self.get_rotation()

        self.get_sprite().position = (pos[0], pos[1])

        x,y = self.get_position()
        R = self.get_rotation_raw()
        rot = [R[0], R[3], R[6], 0.,
               R[1], R[4], R[7], 0.,
               R[2], R[5], R[8], 0.,
               0.0, 0.0, 0.0, 1.0]

        self.get_sprite().draw(mat=rot, bounds_check=bounds_check, colors=colors)


class Geom(physsim.DynamicObject, drawable):

    dynamics_bound = 'box'
    dynamics_size = (1.0, 1.0)
    step_me = False

    collide = True

    dyn_material = {'bounce': 0.0, 'min_bounce': 0.0, 'friction': INFINITY, 'density':INFINITY }
    type = 'geometry'

    def __init__(self, renderlib, pos, size=(1.0, 1.0), graphics_filename='moo.png', tiles=(1.0, 1.0)):
        physsim.DynamicObject.__init__(self, pos=pos)

        self.dynamics_size = size
        self.graphics_size = size
        self.graphics_filenames = [graphics_filename]
        self.current_sprite = graphics_filename
        self.graphics_tiles = tiles

        self.repeats = ('clamp' if self.graphics_tiles[0] == 1.0 else 'repeat',
                        'clamp' if self.graphics_tiles[1] == 1.0 else 'repeat'
                       )

        drawable.__init__(self, renderlib)


    def draw(self, bounds_check=True, colors=True):
        """ draws the prop """
        pos = self.get_position()
        self.get_sprite().position = (pos[0], pos[1])

        x,y = self.get_position()

        self.get_sprite().draw(bounds_check=bounds_check, colors=colors)


class Geom_nodetect(drawable):
    step_me = False
    type = 'geometry_nodetect'

    position = (0.0, 0.0)

    def __init__(self, renderlib, pos, size=(1.0, 1.0), graphics_filename='moo.png', tiles=(1.0, 1.0)):

        self.dynamics_size = size
        self.graphics_size = size
        self.graphics_filenames = [graphics_filename]
        self.current_sprite = graphics_filename
        self.graphics_tiles = tiles

        self.repeats = ('clamp' if self.graphics_tiles[0] == 1.0 else 'repeat',
                        'clamp' if self.graphics_tiles[1] == 1.0 else 'repeat'
                       )

        drawable.__init__(self, renderlib)

    def set_position(self, x, y):
        self.position = (x, y)

    def get_position(self):
        return self.position

    def get_rotation(self):
        return 0

    def get_rotation_raw(self):
        return [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]

    def set_rotation(self, angle):
        """ sets the rotation of the dynamic body """
        raise Exception, 'not implimentated dyn_set_rotation'

    def draw(self, bounds_check=True, colors=True):
        """ draws the prop """
        pos = self.get_position()
        self.get_sprite().position = (pos[0], pos[1])

        x,y = self.get_position()

        self.get_sprite().draw(bounds_check=bounds_check, colors=colors)

#------------------------------------------------------------------------------
# Props
#------------------------------------------------------------------------------
class Prop(physsim.DynamicBody, drawable):


    dynamics_bound = 'box'
    dynamics_size = (1.0, 1.0)
    step_me = False

    dyn_material = { 'bounce':0.01, 'min_bounce':1.0, 'friction':2000, 'density':750 }
    type = 'UNIDENFIEDFEDEDFEDMOOANDBAH'

    def __init__(self, renderlib, pos, override_size=None, id=None):

        if override_size:
            self.dynamics_size = override_size
            self.graphics_size = override_size

        physsim.DynamicBody.__init__(self, pos=pos)
        drawable.__init__(self, renderlib)

class prop_small_box(Prop):

    dynamics_bound = 'box'
    dynamics_size = (0.5, 0.5)

    graphics_filenames = ['props/small_box.png']
    current_sprite = 'props/small_box.png'
    graphics_size = (0.5, 0.5)

    type = 'prop_small_box'

    def __init__(self, *args, **kwargs):
        Prop.__init__(self, *args, **kwargs)

class prop_small_ball(Prop):

    dynamics_bound = 'sphere'
    dynamics_size = (1, 1)

    type = 'prop_small_ball'

    graphics_filenames = ['props/ball.png']
    current_sprite = 'props/ball.png'
    graphics_size = (0.5, 0.5)

    def __init__(self, *args, **kwargs):
        if kwargs.has_key('id'):
            self.logic_id = kwargs['id']
        Prop.__init__(self, *args, **kwargs)

class prop_plank(Prop):
    dynamics_bound = 'box'
    dynamics_size = (0.1, 4.0)

    graphics_filenames = ['props/plank.png']
    current_sprite = 'props/plank.png'
    graphics_size = (1.0, 4.0)

    dyn_material = { 'bounce':0.00, 'min_bounce':0.0, 'friction':10000, 'density':100 }

    type = 'prop_plank'
    islift = True

    def __init__(self, *args, **kwargs):
        Prop.__init__(self, *args, **kwargs)

        self.connect('collide', self.cb_test)

    def cb_test(self, ref, object, info):
        if hasattr(object, 'type'):
            if object.type == 'uppity_box':
                object.disable_me = True

class prop_break_plank(prop_plank):

    step_me = True
    num_lives = 1
    hold_length = 3.0

    dyn_material = { 'bounce':0.00, 'min_bounce':0.0, 'friction':10000, 'density':1000 }

    def __init__(self, *args, **kwargs):

        prop_plank.__init__(self, *args, **kwargs)
        self.timeline = Timeline()
        self.disabled_object = None

        self.original_rot = None
        self.original_pos = None
        self.pos_not_set = True
        self.released = True

        self.sound = instance.instance.audio.create_sound('49139__smidoid__cracking.ogg')

    def create_self(self, *args, **kwargs):
        prop_plank.create_self(self, *args, **kwargs)
        self.original_rot = self.dynamics_geom.getRotation()
        self.original_pos = self.get_position()
        print self.original_pos

    def cb_test(self, ref, object, info):
        if hasattr(object, 'type'):
            if object.type == 'uppity_box':
                self.disabled_object = object
                object.disable_me = True
                #self, time=0.0, relative=time_start, callback=None,
                   #repeat_action=False, repeat_callback=False, length=0.0,
                   #wait=0.0, extra_args={}):
                self.timeline.add_action(time=0.0, relative=time_now,
                                         callback=self.release, length=self.hold_length,
                                         repeat_callback=True, repeat_action=False)
                self.released = False

    def release(self, progress):
        if progress >= 1.0 and not self.released:
            self.released = True
            self.set_position(*self.original_pos)
            self.dynamics_geom.setRotation(self.original_rot)
            self.disabled_object.disable_me = False
            #self.disabled_object.action(True)
            instance.instance.audio.play(self.sound)

    def step(self):
        if self.get_position() is not (0.0, 0.0) and self.pos_not_set:
            self.original_rot = self.dynamics_geom.getRotation()
            self.original_pos = self.get_position()
            self.pos_not_set = False
        self.timeline.step()
